# yellowbrick.gridsearch.pcolor
# Colorplot visualizer for gridsearch results.
#
# Author:   Phillip Schafer
# Created:  Sat Feb 3 10:18:33 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: pcolor.py [03724ed] pbs929@users.noreply.github.com $

"""
Colorplot visualizer for gridsearch results.
"""

import numpy as np

from .base import GridSearchVisualizer


## Packages for export
__all__ = ["GridSearchColorPlot", "gridsearch_color_plot"]


##########################################################################
## Quick method
##########################################################################


def gridsearch_color_plot(estimator, x_param, y_param, X=None, y=None, ax=None, **kwargs):
    """Quick method:
    Create a color plot showing the best grid search scores across two
    parameters.

    This helper function is a quick wrapper to utilize GridSearchColorPlot
    for one-off analysis.

    If no `X` data is passed, the model is assumed to be fit already. This
    allows quick exploration without waiting for the grid search to re-run.

    Parameters
    ----------
    estimator : Scikit-Learn grid search object
        Should be an instance of GridSearchCV. If not, an exception is raised.
        The model may be fit or unfit.

    x_param : string
        The name of the parameter to be visualized on the horizontal axis.

    y_param : string
        The name of the parameter to be visualized on the vertical axis.

    metric : string (default 'mean_test_score')
        The field from the grid search's `cv_results` that we want to display.

    X  : ndarray or DataFrame of shape n x m or None (default None)
        A matrix of n instances with m features. If not None, forces the
        GridSearchCV object to be fit.

    y  : ndarray or Series of length n or None (default None)
        An array or series of target or class values.

    ax : matplotlib axes
        The axes to plot the figure on.

    classes : list of strings
        The names of the classes in the target

    Returns
    -------
    ax : matplotlib axes
        Returns the axes that the classification report was drawn on.
    """
    # Instantiate the visualizer
    visualizer = GridSearchColorPlot(estimator, x_param, y_param, ax=ax, **kwargs)

    # Fit if necessary
    if X is not None:
        visualizer.fit(X, y)
    else:
        visualizer.draw()

    # Return the axes object on the visualizer
    return visualizer.ax


class GridSearchColorPlot(GridSearchVisualizer):
    """
    Create a color plot showing the best grid search scores across two
    parameters.

    Parameters
    ----------
    estimator : Scikit-Learn grid search object
        Should be an instance of GridSearchCV. If not, an exception is raised.

    x_param : string
        The name of the parameter to be visualized on the horizontal axis.

    y_param : string
        The name of the parameter to be visualized on the vertical axis.

    metric : string (default 'mean_test_score')
        The field from the grid search's `cv_results` that we want to display.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    colormap : string or cmap, default: 'RdBu_r'
        optional string or matplotlib cmap to colorize lines
        Use either color to colorize the lines on a per class basis or
        colormap to color them on a continuous scale.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Examples
    --------
    >>> from yellowbrick.gridsearch import GridSearchColorPlot
    >>> from sklearn.model_selection import GridSearchCV
    >>> from sklearn.svm import SVC
    >>> gridsearch = GridSearchCV(SVC(),
                                  {'kernel': ['rbf', 'linear'], 'C': [1, 10]})
    >>> model = GridSearchColorPlot(gridsearch, x_param='kernel', y_param='C')
    >>> model.fit(X)
    >>> model.show()
    """

    def __init__(
        self,
        estimator,
        x_param,
        y_param,
        metric="mean_test_score",
        colormap="RdBu_r",
        ax=None,
        **kwargs
    ):
        super(GridSearchColorPlot, self).__init__(estimator, ax=ax, **kwargs)
        self.x_param = x_param
        self.y_param = y_param
        self.metric = metric
        self.colormap = colormap

    def draw(self):
        # Project the grid search results to 2 dimensions
        x_vals, y_vals, best_scores = self.param_projection(
            self.x_param, self.y_param, metric=self.metric
        )

        # Mask nans so that they can be filled with a hatch
        data = np.ma.masked_invalid(best_scores)

        # Plot and fill in hatch for nans
        mesh = self.ax.pcolor(
            data, cmap=self.colormap, vmin=np.nanmin(data), vmax=np.nanmax(data)
        )
        self.ax.patch.set(hatch="x", edgecolor="black")

        # Ticks and tick labels
        self.ax.set_xticks(np.arange(len(x_vals)) + 0.5)
        self.ax.set_yticks(np.arange(len(y_vals)) + 0.5)
        self.ax.set_xticklabels(x_vals, rotation=45)
        self.ax.set_yticklabels(y_vals, rotation=45)

        # Add the colorbar
        cb = self.ax.figure.colorbar(mesh, None, self.ax)
        cb.outline.set_linewidth(0)

        self.ax.set_aspect("equal")

    def finalize(self):
        self.set_title("Grid Search Scores")
        self.ax.set_xlabel(self.x_param)
        self.ax.set_ylabel(self.y_param)



# yellowbrick.bestfit
# Uses Scikit-Learn to compute a best fit function, then draws it in the plot.
#
# Author:   Benjamin Bengfort
# Created:  Sun Jun 26 17:27:08 2016 -0400
#
# Copyright (C) 2016 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: bestfit.py [56236f3] benjamin@bengfort.com $

"""
Uses Scikit-Learn to compute a best fit function, then draws it in the plot.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.pyplot as plt

from sklearn import linear_model
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import make_pipeline
from sklearn.metrics import mean_squared_error as mse

from operator import itemgetter
from yellowbrick.style.palettes import LINE_COLOR
from yellowbrick.exceptions import YellowbrickValueError


##########################################################################
## Module Constants
##########################################################################

# Names of the various estimator functions
LINEAR = "linear"
QUADRATIC = "quadratic"
EXPONENTIAL = "exponential"
LOG = "log"
SELECT_BEST = "select_best"


##########################################################################
## Draw Line of Best Fit
##########################################################################


def draw_best_fit(X, y, ax, estimator="linear", **kwargs):
    """
    Uses Scikit-Learn to fit a model to X and y then uses the resulting model
    to predict the curve based on the X values. This curve is drawn to the ax
    (matplotlib axis) which must be passed as the third variable.

    The estimator function can be one of the following:

    - ``'linear'``:      Uses OLS to fit the regression
    - ``'quadratic'``:   Uses OLS with Polynomial order 2
    - ``'exponential'``: Not implemented yet
    - ``'log'``:         Not implemented yet
    - ``'select_best'``: Selects the best fit via MSE

    The remaining keyword arguments are passed to ax.plot to define and
    describe the line of best fit.

    Parameters
    ----------
    X : ndarray or DataFrame of shape n x m
        A matrix of n instances with m features

    y : ndarray or Series of length n
        An array or series of target or class values

    ax : matplotlib Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    estimator : string, default: 'linear'
        The name of the estimator function used to draw the best fit line.
        The estimator can currently be one of linear, quadratic, exponential,
        log, or select_best. The select best method uses the minimum MSE to
        select the best fit line.

    kwargs : dict
        Keyword arguments to pass to the matplotlib plot function to style and
        label the line of best fit. By default, the standard line color is
        used unless the color keyword argument is passed in.

    Returns
    -------

    ax : matplotlib Axes
        The axes with the line drawn on it.
    """

    # Estimators are the types of best fit lines that can be drawn.
    estimators = {
        LINEAR: fit_linear,  # Uses OLS to fit the regression
        QUADRATIC: fit_quadratic,  # Uses OLS with Polynomial order 2
        EXPONENTIAL: fit_exponential,  # Not implemented yet
        LOG: fit_log,  # Not implemented yet
        SELECT_BEST: fit_select_best,  # Selects the best fit via MSE
    }

    # Check to make sure that a correct estimator value was passed in.
    if estimator not in estimators:
        raise YellowbrickValueError(
            "'{}' not a valid type of estimator; choose from {}".format(
                estimator, ", ".join(estimators.keys())
            )
        )

    # Then collect the estimator function from the mapping.
    estimator = estimators[estimator]

    # Ensure that X and y are the same length
    if len(X) != len(y):
        raise YellowbrickValueError(
            (
                "X and y must have same length:" " X len {} doesn't match y len {}!"
            ).format(len(X), len(y))
        )

    # Ensure that X and y are np.arrays
    X = np.array(X)
    y = np.array(y)

    # Verify that X is a two dimensional array for Scikit-Learn esitmators
    # and that its dimensions are (n, 1) where n is the number of rows.
    if X.ndim < 2:
        X = X[:, np.newaxis]  # Reshape X into the correct dimensions

    if X.ndim > 2:
        raise YellowbrickValueError(
            "X must be a (1,) or (n,1) dimensional array not {}".format(X.shape)
        )

    # Verify that y is a (n,) dimensional array
    if y.ndim > 1:
        raise YellowbrickValueError(
            "y must be a (1,) dimensional array not {}".format(y.shape)
        )

    # Uses the estimator to fit the data and get the model back.
    model = estimator(X, y)

    # Set the color if not passed in.
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = LINE_COLOR

    # Get the current working axes
    ax = ax or plt.gca()

    # Plot line of best fit onto the axes that were passed in.
    # TODO: determine if xlim or X.min(), X.max() are better params
    xr = np.linspace(*ax.get_xlim(), num=100)
    ax.plot(xr, model.predict(xr[:, np.newaxis]), **kwargs)
    return ax


##########################################################################
## Estimator Functions
##########################################################################


def fit_select_best(X, y):
    """
    Selects the best fit of the estimators already implemented by choosing the
    model with the smallest mean square error metric for the trained values.
    """
    models = [fit(X, y) for fit in [fit_linear, fit_quadratic]]
    errors = map(lambda model: mse(y, model.predict(X)), models)

    return min(zip(models, errors), key=itemgetter(1))[0]


def fit_linear(X, y):
    """
    Uses OLS to fit the regression.
    """
    model = linear_model.LinearRegression()
    model.fit(X, y)
    return model


def fit_quadratic(X, y):
    """
    Uses OLS with Polynomial order 2.
    """
    model = make_pipeline(PolynomialFeatures(2), linear_model.LinearRegression())
    model.fit(X, y)
    return model


def fit_exponential(X, y):
    """
    Fits an exponential curve to the data.
    """
    raise NotImplementedError("Exponential best fit lines are not implemented")


def fit_log(X, y):
    """
    Fit a logrithmic curve to the data.
    """
    raise NotImplementedError("Logrithmic best fit lines are not implemented")


##########################################################################
## Draw 45 Degree Line
##########################################################################


def draw_identity_line(ax=None, dynamic=True, **kwargs):
    """
    Draws a 45 degree identity line such that y=x for all points within the
    given axes x and y limits. This function also registeres a callback so
    that as the figure is modified, the axes are updated and the line remains
    drawn correctly.

    Parameters
    ----------

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    dynamic : bool, default : True
        If the plot is dynamic, callbacks will be registered to update the
        identiy line as axes are changed.

    kwargs : dict
        Keyword arguments to pass to the matplotlib plot function to style the
        identity line.


    Returns
    -------

    ax : matplotlib Axes
        The axes with the line drawn on it.

    Notes
    -----

    .. seealso:: `StackOverflow discussion: Does matplotlib have a function for drawing diagonal lines in axis coordinates? <https://stackoverflow.com/questions/22104256/does-matplotlib-have-a-function-for-drawing-diagonal-lines-in-axis-coordinates>`_
    """

    # Get the current working axes
    ax = ax or plt.gca()

    # Define the standard line color
    if "c" not in kwargs and "color" not in kwargs:
        kwargs["color"] = LINE_COLOR

    # Define the standard opacity
    if "alpha" not in kwargs:
        kwargs["alpha"] = 0.5

    # Draw the identity line
    identity, = ax.plot([], [], **kwargs)

    # Define the callback
    def callback(ax):
        # Get the x and y limits on the axes
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        # Set the bounding range of the line
        data = (max(xlim[0], ylim[0]), min(xlim[1], ylim[1]))
        identity.set_data(data, data)

    # Register the callback and return
    callback(ax)

    if dynamic:
        ax.callbacks.connect("xlim_changed", callback)
        ax.callbacks.connect("ylim_changed", callback)

    return ax


if __name__ == "__main__":
    import os
    import pandas as pd

    path = os.path.join(
        os.path.dirname(__file__), "..", "examples", "data", "concrete.xls"
    )
    if not os.path.exists(path):
        raise Exception("Could not find path for testing")

    xkey = "Fine Aggregate (component 7)(kg in a m^3 mixture)"
    ykey = "Coarse Aggregate  (component 6)(kg in a m^3 mixture)"
    data = pd.read_excel(path)

    fig, axe = plt.subplots()
    axe.scatter(data[xkey], data[ykey])
    draw_best_fit(data[xkey], data[ykey], axe, "select_best")

    plt.show()



# yellowbrick.utils.timer
# Timer utilities
#
# Author:   ZJ Poh
# Created:  Mon Jul 16 10:51:13 2017 -0700
#
# Copyright (C) 2017 The sckit-yb developers
# For license information, see LICENSE.txt
#
# ID: timer.py [75b0f6a] 8103276+zjpoh@users.noreply.github.com $
"""
Timer utilities
"""

##########################################################################
## Imports
##########################################################################

import time

##########################################################################
## Timer Class
##########################################################################


def human_readable_time(s):
    h, s = divmod(s, 3600)
    m, s = divmod(s, 60)
    return "{:>02.0f}:{:02.0f}:{:>07.4f}".format(h, m, s)


class Timer:
    """
    A context object timer. Usage:
        >>> with Timer() as timer:
        ...     do_something()
        >>> print(timer.interval)
    """

    def __init__(self):
        self.time = time.time

    def __enter__(self):
        self.start = self.time()
        return self

    def __exit__(self, *exc):
        self.finish = self.time()
        self.interval = self.finish - self.start

    def __str__(self):
        return human_readable_time(self.interval)



print("import: 'yellowbrick'")
import yellowbrick

print("import: 'yellowbrick.classifier'")
import yellowbrick.classifier

print("import: 'yellowbrick.cluster'")
import yellowbrick.cluster

print("import: 'yellowbrick.contrib'")
import yellowbrick.contrib

print("import: 'yellowbrick.contrib.classifier'")
import yellowbrick.contrib.classifier

print("import: 'yellowbrick.contrib.missing'")
import yellowbrick.contrib.missing

print("import: 'yellowbrick.contrib.statsmodels'")
import yellowbrick.contrib.statsmodels

print("import: 'yellowbrick.datasets'")
import yellowbrick.datasets

print("import: 'yellowbrick.features'")
import yellowbrick.features

print("import: 'yellowbrick.gridsearch'")
import yellowbrick.gridsearch

print("import: 'yellowbrick.model_selection'")
import yellowbrick.model_selection

print("import: 'yellowbrick.regressor'")
import yellowbrick.regressor

print("import: 'yellowbrick.style'")
import yellowbrick.style

print("import: 'yellowbrick.target'")
import yellowbrick.target

print("import: 'yellowbrick.text'")
import yellowbrick.text

print("import: 'yellowbrick.utils'")
import yellowbrick.utils




# yellowbrick.features.base
# Base classes for feature visualizers and feature selection tools.
#
# Author:   Rebecca Bilbro
# Author:   Benjamin Bengfort
# Created:  Fri Oct 07 13:41:24 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: base.py [2e898a6] benjamin@bengfort.com $

"""
Base classes for feature visualizers and feature selection tools.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib as mpl

from yellowbrick.base import Visualizer
from yellowbrick.utils import is_dataframe
from yellowbrick.style import resolve_colors
from yellowbrick.exceptions import NotFitted
from yellowbrick.utils.target import target_color_type, TargetType
from yellowbrick.exceptions import YellowbrickKeyError, YellowbrickValueError
from yellowbrick.style import palettes

from matplotlib.colors import Normalize
from sklearn.base import TransformerMixin


##########################################################################
## Feature Visualizers
##########################################################################


class FeatureVisualizer(Visualizer, TransformerMixin):
    """Base class for feature visualization.

    Feature engineering is primarily conceptualized as a transformation or
    extraction operation, e.g. some raw data is passed through a series of
    transformers and mappings to result in some final dataset which can be
    directly fitted to a model. Therefore feature visualizers are
    transformers and support the sklearn transformer interface by implementing
    a transform method.

    Subclasses of the FeatureVisualizer may call draw either from fit or from
    transform but must implement both so that they can be supported in pipeline
    objects. By default, the transform method of the visualizer is just a data
    pass through that ensures the visualizer can be placed into a feature
    extraction workflow.

    Parameters
    ----------
    ax : matplotlib.Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    kwargs : dict
        Any additional keyword arguments to pass to the base Visualizer.
    """

    def __init__(self, ax=None, fig=None, **kwargs):
        super(FeatureVisualizer, self).__init__(ax=ax, fig=fig, **kwargs)

    def transform(self, X, y=None):
        """
        A pass-through to ensure that feature visualizers work in Pipelines.
        Subclasses may override this method to actually transform data or to
        call drawing methods. The transformer may also take an optional y
        argument if it is required for either transformation or drawing.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature dataset to be transformed.

        y : array-like, shape (n_samples,)
            Dependent target data associated with X, unused.

        Returns
        -------
        X : array-like, shape (n_samples, n_features)
            Returns the original dataset,  unmodified.
        """
        return X

    def fit_transform_show(self, X, y=None, **kwargs):
        """Fit, transform, then visualize data in one step.

        A helper method similar to ``fit_transform`` that allows you to fit,
        transform, and create a visualization in one simple step. Returns a
        transformed dataset similar to ``fit_transform``.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature dataset for both training and transformation.

        y : array-like, shape (n_samples,)
            Dependent target dataset optionally used for training.

        kwargs : dict, optional
            Keyword arguments to pass to the ``show()`` method.

        Returns
        -------
        Xp : array-like, shape (m_samples, m_features)
            The transformed dataset X prime.
        """
        Xp = self.fit_transform(X, y)
        self.show(**kwargs)
        return Xp


class MultiFeatureVisualizer(FeatureVisualizer):
    """Direct visualization of a feature set.

    MultiFeatureVisualiers visualize several features at once, usually in order
    to compare the effectiveness of a subset of features to the superset. This
    type of visualizer provides base functionality for identifying the names of
    the features either directly from the data or from user supplied values. It
    also provides other functionality for feature selection, e.g. ensuring that
    a subset of features is used if specified by the user.

    Parameters
    ----------
    ax : matplotlib.Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    features : list, default: None
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    kwargs : dict, optional
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : ndarray, shape (n_features,)
        The names of the features discovered or used in the visualizer that
        can be used as an index to access or modify data in X. If a user passes
        feature names in, those features are used. Otherwise the columns of a
        DataFrame are used or just simply the indices of the data array.
    """

    def __init__(self, ax=None, fig=None, features=None, **kwargs):
        super(MultiFeatureVisualizer, self).__init__(ax=ax, fig=fig, **kwargs)

        # Data Parameters
        self.features = features

    def fit(self, X, y=None):
        """
        This method performs preliminary computations in order to set up the
        figure or perform other analyses. It can also call drawing methods in
        order to set up various non-instance related figure elements.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature dataset to be transformed.

        y : array-like, shape (n_samples,)
            Optional dependent target data associated with X.

        Returns
        -------
        self : MultiFeatureVisualizer
            Returns the visualizer/transformer for use in Pipelines and chaining.
        """
        n_columns = X.shape[1]

        if self.features is not None:
            # Use the user-specified features with some checking
            # TODO: allow the user specified features to filter the dataset
            if len(self.features) != n_columns:
                raise YellowbrickValueError(
                    (
                        "number of supplied feature names does not match the number "
                        "of columns in the training data."
                    )
                )

            self.features_ = np.array(self.features)

        else:
            # Attempt to determine the feature names from the input data
            if is_dataframe(X):
                self.features_ = np.array(X.columns)

            # Otherwise create numeric labels for each column.
            else:
                self.features_ = np.arange(0, n_columns)

        # Ensure super is called and fit is returned
        super(MultiFeatureVisualizer, self).fit(X, y)
        return self


##########################################################################
## Data Visualizers
##########################################################################


class DataVisualizer(MultiFeatureVisualizer):
    """Visualizations of instances in feature space.

    Data Visualizers plot instances in feature space (sometimes also referred
    to as data space). Feature space is a multi-dimensional space defined by
    the columns of the dataset ``X`` when passed to ``fit()`` and ``transform``.
    These instances and their features are directly plotted in a representation
    of the higher dimensional space.

    Instances can also be labeled by an target vectory, ``y``. The target is
    visualized in data space by color. For example a discrete target for
    classification problems will use categorical colors and a legend. A
    continuous target for regression problems will use sequential colors with
    a colormap.

    This class provides helper functionality related to target identification:
    whether or not the target is sequential or categorical, and mapping a
    color sequence or color set to the targets as appropriate. It also
    determines the scope of the target, e.g. the unique classes or the range
    of the dataset for use in specific visualizations.

    Parameters
    ----------
    ax : matplotlib.Axes, default: None
        The axis to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    fig : matplotlib Figure, default: None
        The figure to plot the Visualizer on. If None is passed in the current
        plot will be used (or generated if required).

    features : list, default: None
        The names of the features specified by the columns of the input dataset.
        This length of this list must match the number of columns in X, otherwise
        an exception will be raised on ``fit()``.

    classes : list, default: None
        The class labels for each class in y, ordered by sorted class index. These
        names act as a label encoder for the legend, identifying integer classes
        or renaming string labels. If omitted, the class labels will be taken from
        the unique values in y.

        Note that the length of this list must match the number of unique values in
        y, otherwise an exception is raised. This parameter is only used in the
        discrete target type case and is ignored otherwise.

    colors : list or tuple, default: None
        A single color to plot all instances as or a list of colors to color each
        instance according to its class in the discrete case or as an ordered
        colormap in the sequential case. If not enough colors per class are
        specified then the colors are treated as a cycle.

    colormap : string or cmap, default: None
        The colormap used to create the individual colors. In the discrete case
        it is used to compute the number of colors needed for each class and
        in the continuous case it is used to create a sequential color map based
        on the range of the target.

    target_type : str, default: "auto"
        Specify the type of target as either "discrete" (classes) or "continuous"
        (real numbers, usually for regression). If "auto", then it will
        attempt to determine the type by counting the number of unique values.

        If the target is discrete, the colors are returned as a dict with classes
        being the keys. If continuous the colors will be list having value of
        color for each point. In either case, if no target is specified, then
        color will be specified as the first color in the color cycle.

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    features_ : ndarray, shape (n_features,)
        The names of the features discovered or used in the visualizer that
        can be used as an index to access or modify data in X. If a user passes
        feature names in, those features are used. Otherwise the columns of a
        DataFrame are used or just simply the indices of the data array.

    classes_ : ndarray, shape (n_classes,)
        The class labels that define the discrete values in the target. Only
        available if the target type is discrete. This is guaranteed to be
        strings even if the classes are a different type.

    range_ : (min y, max y)
        A tuple that describes the minimum and maximum values in the target.
        Only available if the target type is continuous.
    """

    def __init__(
        self,
        ax=None,
        fig=None,
        features=None,
        classes=None,
        colors=None,
        colormap=None,
        target_type="auto",
        **kwargs
    ):
        super(DataVisualizer, self).__init__(
            ax=ax, fig=fig, features=features, **kwargs
        )

        # Validate raises YellowbrickValueError if invalid
        TargetType.validate(target_type)

        # Data Parameters
        self.classes = classes
        self.target_type = target_type

        # Visual Parameters
        self.colors = colors
        self.colormap = colormap

        # Internal attributes
        self._colors = None
        self._target_color_type = None
        self._label_encoder = None

    def fit(self, X, y=None):
        """
        Fits the visualizer to the training data set by determining the
        target type, colors, classes, and range of the data to ensure that
        the visualizer can accurately portray the instances in data space.

        Parameters
        ----------
        X : ndarray or DataFrame of shape n x m
            A matrix of n instances with m features

        y : ndarray or Series of length n
            An array or series of target or class values

        Returns
        -------
        self : DataVisualizer
            Returns the instance of the transformer/visualizer
        """
        # Compute the features from the data
        super(DataVisualizer, self).fit(X, y)

        # Determine the target color type
        self._determine_target_color_type(y)

        # Handle the single target color type
        if self._target_color_type == TargetType.SINGLE:
            # use the user supplied color or the first color in the color cycle
            self._colors = self.colors or "C0"

        # Compute classes and colors if target type is discrete
        elif self._target_color_type == TargetType.DISCRETE:
            # Unique labels are used both for validation and color mapping
            labels = np.unique(y)

            # Handle user supplied classes
            if self.classes is not None:
                self.classes_ = np.asarray([str(c) for c in self.classes])

                # Validate user supplied class labels
                if len(self.classes_) != len(labels):
                    raise YellowbrickValueError(
                        (
                            "number of specified classes does not match "
                            "number of unique values in target"
                        )
                    )

            # Get the string labels from the unique values in y
            else:
                self.classes_ = np.asarray([str(c) for c in labels])

            # Create a map of class labels to colors
            color_values = resolve_colors(
                n_colors=len(self.classes_), colormap=self.colormap, colors=self.colors
            )
            self._colors = dict(zip(self.classes_, color_values))
            self._label_encoder = dict(zip(labels, self.classes_))

        # Compute target range if colors are continuous
        elif self._target_color_type == TargetType.CONTINUOUS:
            y = np.asarray(y)
            self.range_ = (y.min(), y.max())
            if self.colormap is None:
                self.colormap = palettes.DEFAULT_SEQUENCE
            # TODO: allow for Yellowbrick palettes here as well
            self._colors = mpl.cm.get_cmap(self.colormap)

        # If this exception is raised a developer error has occurred because
        # unknown types should have errored when the type was determined.
        else:
            raise YellowbrickValueError(
                "unknown target color type '{}'".format(self._target_color_type)
            )

        # NOTE: cannot call draw in fit to support data transformers
        return self

    def _determine_target_color_type(self, y):
        """
        Determines the target color type from the vector y as follows:

            - if y is None: only a single color is used
            - if target is auto: determine if y is continuous or discrete
            - otherwise specify supplied target type

        This property will be used to compute the colors for each point.
        """
        if y is None:
            self._target_color_type = TargetType.SINGLE
        elif self.target_type == TargetType.AUTO:
            self._target_color_type = target_color_type(y)
        else:
            self._target_color_type = TargetType(self.target_type)

        # Ensures that target is either SINGLE, DISCRETE or CONTINUOUS before continuing
        if (
            self._target_color_type == TargetType.AUTO
            or self._target_color_type == TargetType.UNKNOWN
        ):
            raise YellowbrickValueError(
                (
                    "could not determine target color type " "from target='{}' to '{}'"
                ).format(self.target_type, self._target_color_type)
            )

    def get_target_color_type(self):
        """
        Returns the computed target color type if fitted or specified by the user.
        """
        if self._target_color_type is None:
            raise NotFitted("unknown target color type on unfitted visualizer")
        return self._target_color_type

    def get_colors(self, y):
        """
        Returns the color for the specified value(s) of y based on the learned
        colors property for any specified target type.

        Parameters
        ----------
        y : array-like
            The values of y to get the associated colors for.

        Returns
        -------
        colors : list
            Returns a list of colors for each value in y.
        """
        if self._colors is None:
            raise NotFitted("cannot determine colors on unfitted visualizer")

        if self._target_color_type == TargetType.SINGLE:
            return [self._colors] * len(y)

        if self._target_color_type == TargetType.DISCRETE:
            try:
                # Use the label encoder to get the class name (or use the value
                # if the label is not mapped in the encoder) then use the class
                # name to get the color from the color map.
                return [self._colors[self._label_encoder.get(yi, yi)] for yi in y]
            except KeyError:
                unknown = set(y) - set(self._label_encoder.keys())
                unknown = ", ".join(["'{}'".format(uk) for uk in unknown])
                raise YellowbrickKeyError(
                    "could not determine color for classes {}".format(unknown)
                )

        if self._target_color_type == TargetType.CONTINUOUS:
            # Normalize values into target range and compute colors from colormap
            norm = Normalize(*self.range_)
            return self._colors(norm(y))

        # This is a developer error, we should never get here!
        raise YellowbrickValueError(
            "unknown target color type '{}'".format(self._target_color_type)
        )



# yellowbrick.style
# Manage the style and aesthetic of the yellowbrick library.
#
# Author:   Benjamin Bengfort
# Created:  Tue Oct 04 15:09:48 2016 -0400
#
# Copyright (C) 2016 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: __init__.py [c6aff34] benjamin@bengfort.com $

"""
Manage the style and aesthetic of the yellowbrick library.
"""

##########################################################################
## Imports
##########################################################################

from .colors import *
from .palettes import *
from .rcmod import *
from .utils import *



# yellowbrick.model_selection.cross_validation
# Implements cross-validation score plotting for model selection.
#
# Author:   Prema Damodaran Roman
# Created:  Wed June 6 2018 13:32:00 -0500
# Author:   Rebecca Bilbro
# Updated:  Fri Aug 10 13:15:43 2018 -0500
#
# Copyright (C) 2018 The scikit-yb developers
# For license information, see LICENSE.txt
#
# ID: cross_validation.py [7f47800] pdamo24@gmail.com $

"""
Implements cross-validation score plotting for model selection.
"""

##########################################################################
# Imports
##########################################################################

import numpy as np
import matplotlib.ticker as ticker

from yellowbrick.base import ModelVisualizer
from sklearn.model_selection import cross_val_score


##########################################################################
# CVScores Visualizer
##########################################################################


class CVScores(ModelVisualizer):
    """
    CVScores displays cross-validated scores as a bar chart, with the
    average of the scores plotted as a horizontal line.

    Parameters
    ----------

    estimator : a scikit-learn estimator
        An object that implements ``fit`` and ``predict``, can be a
        classifier, regressor, or clusterer so long as there is also a valid
        associated scoring metric.
        Note that the object is cloned for each validation.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.

        See the scikit-learn `cross-validation guide <https://goo.gl/FS3VU6>`_
        for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

        See scikit-learn `cross-validation guide <https://goo.gl/FS3VU6>`_
        for more information on the possible metrics that can be used.

    color: string
        Specify color for barchart

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    -------
    cv_scores_ : ndarray shape (n_splits, )
        The cross-validated scores from each subsection of the data

    cv_scores_mean_ : float
        Average cross-validated score across all subsections of the data


    Examples
    --------

    >>> from sklearn import datasets, svm
    >>> iris = datasets.load_iris()
    >>> clf = svm.SVC(kernel='linear', C=1)
    >>> X = iris.data
    >>> y = iris.target
    >>> visualizer = CVScores(estimator=clf, cv=5, scoring='f1_macro')
    >>> visualizer.fit(X,y)
    >>> visualizer.show()

    Notes
    -----

    This visualizer is a wrapper for
    `sklearn.model_selection.cross_val_score <https://goo.gl/4v7dfL>`_.

    Refer to the scikit-learn
    `cross-validation guide <https://goo.gl/FS3VU6>`_
    for more details.

    """

    def __init__(self, estimator, ax=None, cv=None, scoring=None, color=None, **kwargs):
        super(CVScores, self).__init__(estimator, ax=ax, **kwargs)

        self.cv = cv
        self.color = color
        self.scoring = scoring

    def fit(self, X, y, **kwargs):
        """
        Fits the learning curve with the wrapped model to the specified data.
        Draws training and test score curves and saves the scores to the
        estimator.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.

        y : array-like, shape (n_samples) or (n_samples, n_features), optional
            Target relative to X for classification or regression;
            None for unsupervised learning.

        Returns
        -------
        self : instance

        """

        self.cv_scores_ = cross_val_score(
            self.estimator, X, y, cv=self.cv, scoring=self.scoring
        )
        self.cv_scores_mean_ = self.cv_scores_.mean()

        self.draw()
        return self

    def draw(self, **kwargs):
        """
        Creates the bar chart of the cross-validated scores generated from the
        fit method and places a dashed horizontal line that represents the
        average value of the scores.
        """

        width = kwargs.pop("width", 0.3)
        linewidth = kwargs.pop("linewidth", 1)

        xvals = np.arange(1, len(self.cv_scores_) + 1, 1)
        self.ax.bar(xvals, self.cv_scores_, width=width, color=self.color)
        self.ax.axhline(
            self.cv_scores_mean_,
            color=self.color,
            label="Mean score = {:0.3f}".format(self.cv_scores_mean_),
            linestyle="--",
            linewidth=linewidth,
        )

        return self.ax

    def finalize(self, **kwargs):
        """
        Add the title, legend, and other visual final touches to the plot.
        """

        # Set the title of the figure
        self.set_title("Cross Validation Scores for {}".format(self.name))

        # Add the legend
        loc = kwargs.pop("loc", "best")
        edgecolor = kwargs.pop("edgecolor", "k")
        self.ax.legend(frameon=True, loc=loc, edgecolor=edgecolor)

        # set spacing between the x ticks
        self.ax.xaxis.set_major_locator(ticker.MultipleLocator(1))

        # Set the axis labels
        self.ax.set_xlabel("Training Instances")
        self.ax.set_ylabel("Score")


##########################################################################
# Quick Method
##########################################################################


def cv_scores(
    estimator, X, y, ax=None, cv=None, scoring=None, color=None, show=True, **kwargs
):
    """
    Displays cross validation scores as a bar chart and the
    average of the scores as a horizontal line

    This helper function is a quick wrapper to utilize the
    CVScores visualizer for one-off analysis.

    Parameters
    ----------

    estimator : a scikit-learn estimator
        An object that implements ``fit`` and ``predict``, can be a
        classifier, regressor, or clusterer so long as there is also a valid
        associated scoring metric.
        Note that the object is cloned for each validation.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression;
        None for unsupervised learning.

    ax : matplotlib.Axes object, optional
        The axes object to plot the figure on.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross-validation,
        - integer, to specify the number of folds.
        - An object to be used as a cross-validation generator.
        - An iterable yielding train/test splits.

    see the scikit-learn
    `cross-validation guide <https://goo.gl/FS3VU6>`_
    for more information on the possible strategies that can be used here.

    scoring : string, callable or None, optional, default: None
        A string or scorer callable object / function with signature
        ``scorer(estimator, X, y)``.

        See scikit-learn `cross-validation guide <https://goo.gl/FS3VU6>`_
        for more information on the possible metrics that can be used.

    color: string
        Specify color for barchart

    show: bool, default: True
        If True, calls ``show()``, which in turn calls ``plt.show()`` however you cannot
        call ``plt.savefig`` from this signature, nor ``clear_figure``. If False, simply
        calls ``finalize()``

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Returns
    -------
    visualizer : CVScores
        The fitted visualizer.

    """

    # Initialize the visualizer
    visualizer = CVScores(
        estimator, ax=ax, cv=cv, scoring=scoring, color=None, **kwargs
    )

    # Fit and show the visualizer
    visualizer.fit(X, y)

    if show:
        visualizer.show()
    else:
        visualizer.finalize()

    # Return the visualizer object
    return visualizer
